//
//    object-c version
//   
//    to do
//      color handling
//      avoiding Xarray calculation 
//      reading text file
//      setup for print/PDF



#import "GraphView.h"


@implementation GraphView

//-------------------------------------------------
//         initialization
//
- (id)initWithFrame:(NSRect)frameRect
{
	self = [super initWithFrame:frameRect];
	if (self == nil) return nil;
    
	isPDFOutput = NO;
	isDragging = NO;	
	dataExists = NO;
    
	colorNumber = 0;
	mPoint = NSMakePoint(0,0);
	X = nil; 
	Y = nil;
	nt=0;
	na=0;
	
//   dataFilePath = [[NSString alloc] init];
	return self;
}


 -(void)awakeFromNib
 {
// NSLog(@"GraphView awakeFromNib");
 
 return;
 }
 

//-------------------------------------------------
//         destructor
//

- (void)dealloc
{
	if(X!=nil) delete [] X;
	if(na>0){
		for(int i=0;i<nt;i++) delete [] Y[i];
	}
	if(Y!=nil) delete [] Y;	
	dataExists = NO;
	[super dealloc];
}


- (void)outputPDFFile:(BOOL)flag withFile:(NSString*)pdfFilename;
{
    isPDFOutput = flag;
    dataFilePath =[[NSString alloc] initWithString:pdfFilename];
//    dataFilePath = [NSString stringWithString:pdfFilename];
}

//-------------------------------------------------
//         messaging functions
//

- (void)setColorNumber:(int)number
{
	colorNumber = number;
}



//=======================================
//      Open file for reading
//  

-(void) readDataFile:(NSString*)filename with:(int)ntraces
{
	FILE *fp;
	int i,j;
	int datasize;
	
	int filetype=0;    // 1: binary file, 2: text file
	BOOL Success;
	double z;
	double *a;
	char b[512];
    
    
	
	// at first, clear buffers if already present 
    if(X!=nil) delete [] X;
	
	if(na>0){
		for(i=0;i<nt;i++) if(Y[i]!=nil) delete [] Y[i];
	}	
	if(Y!=nil) delete [] Y;
	
    if ([filename rangeOfString:@".dat"].location != NSNotFound){
		filetype=1;
	}else if ([filename rangeOfString:@".txt"].location != NSNotFound){
		filetype = 2;
	}
	if(filetype==0)return;
	
	if(filetype==1){   // binary file
		double a0, d0, d1;
		if((fp = fopen((char*)[filename UTF8String], "rb"))==nil) return;				
        // check file size		
		fseek(fp,0, SEEK_END);
		datasize = ftell(fp)/sizeof(double);		
		if(ntraces==0){  // guess the number of traces
			Success = NO;	
			for(na=2 ; na < datasize/4 ; na++){
				if(datasize % na ==0){
					fseek(fp,0,SEEK_SET);
					fread(&a0,sizeof(double), 1, fp);
					fseek(fp,na*sizeof(double),SEEK_SET);
					fread(&d0,sizeof(double), 1, fp);
					d0 = d0 - a0;
					
					if(d0!=0){
						i=2;
						while(1){
							fseek(fp,i*na*sizeof(double),SEEK_SET);
							fread(&d1,sizeof(double), 1, fp);
							d1 = (d1-a0)/(double)i;
							if(fabs((d1-d0)/d0)>1e-3 )break;
							i++;
							if((i*na>=datasize) || (i>30)){
								Success = YES;
								break;
							}
						} // endo of while-loop
					}
				}
				if(Success==YES)break; 
			}  // end of for-loop
			nt = na-1;
			ntraces = nt;
		}else{
			nt=ntraces;			
			na=ntraces+1;
		}
        
	}else if(filetype==2){  // text file
        
		if((fp = fopen((char*)[filename UTF8String], "rt"))==nil) return;				
		
		if(ntraces==0){
			fgets(b,512,fp);			
			char* pb;
			pb=b;
			i=0;
			Success = NO;
			while(*pb!=0){
				if(isblank(*pb++)==0){   
					if(Success==NO){   // blank -> non-blank
						Success = YES;
						i++;
					}
				}else{	
					if(Success==YES)Success=NO; // non-blank -> blank
				}
			}
			na = i;
			nt = i-1;
		}else{
			na=ntraces+1;
			nt=ntraces;
		}
		
		rewind(fp);
		datasize=0;
		while(1){
			if(fscanf(fp, "%lf", &z) == 1) {
				datasize++;
			}else{
				break;
			}
		}
	}
	
	npts = datasize/na;	
	X = new double[npts];   
	Y = new double_ptr[nt]; 
    for(i=0;i<nt;i++){
		Y[i] = new double[npts];    
	}
    
	
	if(filetype==1){ // binary file
		a = new double[na];   
		fseek(fp,0,SEEK_SET);
		for(j=0;j<npts;j++){
			fread(a,sizeof(double), na, fp);
			X[j] = a[0];
			for(i=0;i<nt;i++){
				Y[i][j] = a[i+1];
			}
		}
		fclose(fp);		
		if(a!=nil) delete [] a;
	}else if(filetype==2){   // text file
		rewind(fp);		
		fseek(fp,0,SEEK_SET);
		for(j=0;j<npts;j++){
			fscanf(fp, "%lf", &z);
			X[j] = z;
			for(i=0;i<nt;i++){
				fscanf(fp, "%lf", &z);
				Y[i][j] = z;
			}
		}						
		fclose(fp);
	} 
    
    // Drawing range
	
	IMin = 0;
	IMax = npts-1;
    
    // X axis ragne ( can be non-linear)    
    //     XMin and XMax are not changed if the data are the same
    //     rXMin and rXMax are limits of the displayed graph.
    
    XMin = XMax = X[0];
    for(j=0;j<npts;j++){
        if(XMin > X[j]){
            XMin = X[j];
        }else if (XMax < X[j]){
            XMax = X[j];
        }
    }
    rXMin = XMin;
    rXMax = XMax;
    
    // y axis    
	YMin = YMax = Y[0][0];
	for(i=0;i<nt;i++){
		for(j=0;j<npts;j++){
			if(YMin > Y[i][j]){
				YMin = Y[i][j];
			}else if(YMax < Y[i][j]){
				YMax = Y[i][j];
			}
		}
	}
	if(YMax==YMin){
		YMax+=0.5;
		YMin-=0.5;
	}	
	rYMin = YMin;
	rYMax = YMax;
	
	dataExists=YES;
}





//======================================================
//      Scaling functions
//

-(float) xScale:(double) x
{
	return cXOffset + XScaleFactor*(x-rXMin);
}	

-(float) yScale:(double)  y
{
	return cYOffset + YScaleFactor*(y-rYMin);
}	

-(double)xScaleBack:(int) ix
{
	return  rXMin + (ix-cXOffset)/XScaleFactor;
}


-(double) yRange:(double)y
{
    if(y<rYMin)return rYMin;
    if(y>rYMax)return rYMax;
    return y;
}

-(double) yScaleBack:(int)iy
{
	return  rYMin + (iy - cYOffset)/YScaleFactor;
}


-(double)autoRangeFrom:(double)ZMin To:(double) ZMax
{
	int nTic = 3;
	
	double ZDelta = ZMax-ZMin;
    if(ZDelta<0)ZDelta=-ZDelta;
	double ZTic = (int)log10(ZDelta);
	ZTic = pow(10.0, (double)ZTic);
	while(1){
		if ( (ZDelta/ZTic) >= nTic)break;
		
		ZTic = 0.5 *  ZTic;
		if((ZDelta/ZTic)>= nTic)break;
		
		ZTic = 0.4 *  ZTic;
		if((ZDelta/ZTic)>=nTic)break;
		
		ZTic = 0.5 *  ZTic;
		if((ZDelta/ZTic)>=nTic)break;
	}
	return ZTic;
}



//=================================================
//    Primitive drawing functions
//

-(void) setFont:(char*)fontname Size:(float)fontsize atRotation:(float)rotation withColor:(RGBColorF)c
{
	CGAffineTransform myTextTransform;
	CGContextSelectFont (context, fontname, fontsize, kCGEncodingMacRoman);
	CGContextSetTextDrawingMode (context, kCGTextFill); 
	CGContextSetRGBFillColor (context, c.red, c.green, c.blue, c.alpha);
	myTextTransform = CGAffineTransformMakeRotation (rotation);
	CGContextSetTextMatrix (context, myTextTransform);
}


-(void) textOutAtPoint:(CGPoint)p string:(NSString*)s align:(int)textAlign
{
	CGPoint p0, p1;
	p0.x = p.x;
	p0.y = p.y;
	
	if(textAlign!=textAlignLeft){
		CGContextSetTextDrawingMode (context, kCGTextInvisible); 
		CGContextShowTextAtPoint (context, p.x, p.y, (char*)[s UTF8String], [s length]);
		p1 = CGContextGetTextPosition(context);
		if(textAlign==textAlignCenter){
			p0.x= (3*p0.x - p1.x)/2;
			p0.y=(3*p0.y - p1.y)/2;
		}else if (textAlign==textAlignRight){
			p0.x = 2*p0.x - p1.x; 
			p0.y = 2*p0.y - p1.y; 		
		}
		CGContextSetTextDrawingMode (context, kCGTextFill); 		
	}
	CGContextShowTextAtPoint (context, p0.x, p0.y, (char*)[s UTF8String], [s length]);
}


-(void) drawLineFrom:(CGPoint)p0 To:(CGPoint)p1 withColor:(RGBColorF) c
{
	CGContextBeginPath(context);
	CGContextMoveToPoint(context, p0.x, p0.y);
	CGContextAddLineToPoint(context, p1.x, p1.y);
	CGContextSetRGBStrokeColor(context, c.red, c.green, c.blue, c.alpha);
	CGContextSetLineWidth(context, 1);
	CGContextStrokePath(context);
}



-(void) drawPolyline:(int)itx
{
	int ic;
    
	if(itx<0 || itx>=nt)return;
	RGBColorF color={0.0, 0.0, 0.0, 1.0};
	CGContextBeginPath(context);
	
    float ix0 = [self xScale:X[IMin]];
	double y = Y[itx][IMin];
    
    //	if(y<rYMin){ y = rYMin; }else if(y>rYMax){ y = rYMax;}  
	float iy0 = [self yScale:[self yRange:y]];
	
	CGContextMoveToPoint(context, ix0, iy0);
	for(int j=IMin+1 ; j <= IMax ; j++){
		float ix1 = [self xScale:X[j]];   
		y = Y[itx][j];
        //		if(y<rYMin){ y = rYMin; }else if(y>rYMax){ y = rYMax;}  
		float iy1 = [self yScale:[self yRange:y]];
        
		if(((int)ix1!=(int)ix0) || ((int)iy1!=(int)iy0)){      
			CGContextAddLineToPoint(context, ix1, iy1);
			ix0 = ix1;
			iy0 = iy1;
		}  
	}
	
	switch(colorNumber){
		case 0:
			ic = itx % 8;
			break;
		case 1:
			ic = 0;
			break;
		case 2:
			ic = 1;
			break;
		case 3:
			ic = 2;
			break;
        defalut:
			ic = 10;
	}
	
	color.alpha=1.0;
	switch( ic) {
		case 0:
			color.red=0.86; color.green=0.08; color.blue=0.24; //red
			break;
		case 1:
			color.red=0.13; color.green=0.54; color.blue=0.13; //green
			break;
		case 2:
			color.red=0.28; color.green=0.51; color.blue=0.71;  //blue
			break;
		case 3:
			color.red=0.95; color.green=0.27; color.blue=0.0; // orange
			break;
		case 4:
			color.red=0.0; color.green=0.0; color.blue=0.50;  // dark blue
			break;
		case 5:
			color.red=1.0; color.green=0.84; color.blue=0.0;  //yellow
			break;
		case 6:
			color.red=0.58; color.green=0.0; color.blue=0.83; // purple
			break;
		case 7:
			color.red=0.41; color.green=0.41; color.blue=0.41; //gray
			break;
		default:
			color.red=0.0; color.green=0.0; color.blue=0.0;    //black
			;
	}    
	CGContextSetRGBStrokeColor(context, color.red, color.green, color.blue, color.alpha);
	CGContextSetLineWidth(context, 1);
	CGContextStrokePath(context);
}


//=================================================
//          Draw Graph
//


-(void) drawGraph 
{
	int i, n;
	
	float fontSize;
	//float w, h;
	double TicUnit;
	double aX, aY;
	double TicLength=10;
	NSString* s;
	CGPoint p0, p1;
	RGBColorF color={0.0, 0.0, 0.0, 1.0};
	
	CGContextSetGrayFillColor(context, 1.0, 1.0);
	CGContextFillRect(context, r);
	
    cXOffset = 0.1*w;
	cYOffset = 0.1*h;
	XScaleFactor = 0.85*w/(rXMax-rXMin);        
    YScaleFactor = 0.85*h/(rYMax-rYMin);
    
	if(isPDFOutput){
        fontSize = h/20;
    }else{
        fontSize = h/40;
	}
	color.green=1.0;
	color.alpha=0.5;
	[self setFont:(char*)[@"Helvetica" UTF8String] Size:fontSize atRotation:0 withColor:color];
	if(!isPDFOutput){
        p0.x = mPoint.x;
        p0.y = mPoint.y;	
        s = [NSString stringWithFormat:@"(%lf,%lf)", [self xScaleBack:(int)mPoint.x], [self yScaleBack:(int)mPoint.y] ];
        [self textOutAtPoint:p0 string:s align:textAlignLeft];
    }    
    
	color.red=color.green=color.blue=0.0;
	
	// X-Axis 
	[self setFont:(char*)[@"Helvetica" UTF8String] Size:fontSize atRotation:0 withColor:color];
//	[self setFont:"Helvetica" Size:fontSize atRotation:0 withColor:color];
	p0.x = [self xScale:rXMin];
	p0.y = [self yScale:rYMin];
	p1.x = [self xScale:rXMax];
	p1.y = [self yScale:rYMin];
	[self drawLineFrom:p0 To:p1 withColor:color];
	
	TicUnit = [self autoRangeFrom:rXMin To:rXMax];
    
	n = (int)( -log10(TicUnit)+1 );
	if( n < 0 ) n = 0;	
	
	i= (int)(rXMin / TicUnit);
	if( i*TicUnit < rXMin ) i++;
	while(1){
		aX = i*TicUnit;
		if(aX > rXMax ) break;
		s = [NSString stringWithFormat:@"%.*f",n,aX];
		
		p0.x = [self xScale:aX];
		p0.y = [self yScale:rYMin]-1.5*fontSize;
		[self textOutAtPoint:p0 string:s align:textAlignCenter];
		
		p0.x = p1.x = [self xScale:aX];
		p0.y = [self yScale:rYMin];
		p1.y = p0.y-TicLength;
		
		[self drawLineFrom:p0 To:p1 withColor:color];
		i++;
	}
	
	// Y-Axis 
	
	p0.x = [self xScale:rXMin];
	p0.y = [self yScale:rYMin];
	p1.x = [self xScale:rXMin];
	p1.y = [self yScale:rYMax];	
	[self drawLineFrom:p0 To:p1 withColor:color];
	
	TicUnit = [self autoRangeFrom:rYMin To:rYMax];
	
	n = (int)( -log10(TicUnit)+1 );
	if( n < 0 ) n = 0;	
	
    
	i= (int)(rYMin / TicUnit);
	if( i*TicUnit < rYMin ) i++;
	while(1){
		aY = i*TicUnit;
		if(aY > rYMax ) break;
		s = [NSString stringWithFormat:@"%.*f",n,aY];
		
		p0.x = [self xScale:rXMin] -1.5*TicLength;
		p0.y = [self yScale:aY]-0.33*fontSize;
		
        [self textOutAtPoint:p0 string:s align:textAlignRight];
		
		p0.x = [self xScale:rXMin];
		p1.x = p0.x-TicLength;
		p0.y = p1.y = [self yScale:aY];
		[self drawLineFrom:p0 To:p1 withColor:color];
		
		i++;
	}
	
	//  Draw traces
	
	for(i=0;i<nt;i++){
        [self drawPolyline:i];
	}
	if(!isDragging)return;
	
	// selection rectangle
	
	CGContextMoveToPoint(context, originPoint.x, originPoint.y);
	CGContextAddLineToPoint(context, mPoint.x, originPoint.y);
	CGContextAddLineToPoint(context, mPoint.x, mPoint.y);
	CGContextAddLineToPoint(context, originPoint.x, mPoint.y);
	CGContextAddLineToPoint(context, originPoint.x, originPoint.y);
	
	CGContextSetRGBStrokeColor(context, 0.3, 0.3, 0.3, 1.0);
	CGContextSetLineWidth(context, 0.5);
	CGContextStrokePath(context);
	
}

//****************************************************************
//                    EVENT HANDLERS
//

- (void)drawRect:(NSRect)rect
{
	if(dataExists==NO) return;
    r = CGRectMake(rect.origin.x, rect.origin.y, rect.size.width, rect.size.height);
    w = CGRectGetWidth(r);
    h = CGRectGetHeight(r);
    if(isPDFOutput){
        h = (w!=0)? h*400/w : 300;
        w = 400;
    	r = CGRectMake(0,  0,  w, h);
        CFURLRef url = CFURLCreateWithFileSystemPath(NULL, (CFStringRef)dataFilePath, kCFURLPOSIXPathStyle,false);
        [dataFilePath release];
        if (url == NULL)return;            
        context = CGPDFContextCreateWithURL (url, &r,NULL);
        CFRelease(url);
        CGContextBeginPage(context, &r);
    }else{
        context = (CGContextRef)[[NSGraphicsContext currentContext] graphicsPort];
    }
    [self drawGraph];
    
    if(isPDFOutput){
        CGContextEndPage(context);
        CGContextRelease(context);
        isPDFOutput=NO;
    }
}


//*********************************
//      MOUSE LEFT BUTTON
//
- (void)mouseDown:(NSEvent *)theEvent 
{
    if(dataExists==NO)return;
    if(isPDFOutput)return;
    isDragging = YES;
    originPoint = [self convertPoint:[theEvent locationInWindow] fromView:nil];  
    mPoint = originPoint;
    [self setNeedsDisplay:YES];
}


- (void) mouseDragged:(NSEvent *) theEvent
{
    if(dataExists==NO)return;
    if(isPDFOutput)return;
    isDragging = YES;
    mPoint = [self convertPoint:[theEvent locationInWindow] fromView:nil];  
    
    [self setNeedsDisplay:YES];				
}


- (void)mouseUp:(NSEvent *)theEvent 
{
    bool inRange = false;
    int iXlow;
    int iXup;
    
    if(dataExists==NO)return;
    if(isPDFOutput)return;
    isDragging = NO;
    
    double x0 = [self xScaleBack:(int)originPoint.x];
    double x1 = [self xScaleBack:(int)mPoint.x];
    
    if(x0 > x1){
        double tmpx = x0;
        x0 = x1;
        x1 = tmpx;
    }
    
    for(int i= IMin ; i<=IMax ; i++){
        if((X[i]>=x0) && (X[i]<=x1)){
            if(inRange){
                if(i<iXlow){
                    iXlow = i;
                }else if(i>iXup){
                    iXup = i;
                }
            }else{
                iXlow = iXup = i;
                inRange=true;
            }        
        }
    }
    
    if(inRange){
        IMin = iXlow;
        IMax = iXup;
        if(IMin == IMax){
            IMin = IMin > 0 ? IMin-1 : 0;
            IMax = IMax < npts -1 ? IMax+1 : npts-1;
        }
        rXMin = X[IMin];
        rXMax = X[IMax];
        if(rXMin > rXMax){
            rXMin = X[IMax];
            rXMax = X[IMin];
        }
    }    // if(!inRange) IMin nor IMax are changed
    
    [self setNeedsDisplay:YES];
}

//*********************************
//       RIGHT BUTTON
//
- (void)rightMouseDown:(NSEvent *)theEvent 
{
    if(dataExists==NO)return;
    if(isPDFOutput)return;
    isDragging = YES;
    originPoint = [self convertPoint:[theEvent locationInWindow] fromView:nil];  
    mPoint = originPoint;
    [self setNeedsDisplay:YES];
}


- (void) rightMouseDragged:(NSEvent *) theEvent
{
    if(dataExists==NO)return;
    if(isPDFOutput)return;
    isDragging = YES;
    mPoint = [self convertPoint:[theEvent locationInWindow] fromView:nil];  
    [self setNeedsDisplay:YES];				
}


- (void)rightMouseUp:(NSEvent *)theEvent 
{
    double y0, y1;
    
    if(dataExists==NO)return;
    if(isPDFOutput)return;
    
    isDragging = NO;
    
    y0 = [self yScaleBack:(int)originPoint.y];
    y1 = [self yScaleBack:(int)mPoint.y];
    
    if(y0==y1){
        return;
    }else if(y0<y1){
        rYMin=y0;
        rYMax = y1;
    }else{
        rYMin = y1;
        rYMax = y0;
    }	
    [self setNeedsDisplay:YES];
}


//*********************************
//       KEY BOARD EVENT
//

- (void) keyDown: (NSEvent *) event
{
    if(dataExists==NO)return;
    if(isPDFOutput)return;
    
    if(([event modifierFlags] & NSControlKeyMask)==0)return;
    
    if([[event charactersIgnoringModifiers] characterAtIndex: 0]=='a'){
        IMin = 0;
        IMax = npts-1;
        rXMin = XMin;
        rXMax = XMax;
        
        rYMin = YMin;
        rYMax = YMax;
        [self setNeedsDisplay:YES];
    }
}

//*********************************
//          enable to accept key input
//

-(BOOL)acceptsFirstResponder
{
    return YES;
}

@end
